package sun.misc;

import java.security.AccessController;
import java.security.PrivilegedAction;
import java.security.ProtectionDomain;
import java.util.concurrent.atomic.AtomicInteger;
import java.security.AccessControlContext;

public final class InnocuousThread extends Thread
{
    private static final Unsafe UNSAFE;
    private static final long THREAD_LOCALS;
    private static final long INHERITABLE_THREAD_LOCALS;
    private static final ThreadGroup INNOCUOUSTHREADGROUP;
    private static final AccessControlContext ACC;
    private static final long INHERITEDACCESSCONTROLCONTEXT;
    private static final long CONTEXTCLASSLOADER;
    private static final AtomicInteger threadNumber;
    private volatile boolean hasRun;
    
    private static String newName() {
        return "InnocuousThread-" + InnocuousThread.threadNumber.getAndIncrement();
    }
    
    public static Thread newSystemThread(Runnable runnable) {
        return newSystemThread(newName(), runnable);
    }
    
    public static Thread newSystemThread(String s, Runnable runnable) {
        return new InnocuousThread(InnocuousThread.INNOCUOUSTHREADGROUP, runnable, s, null);
    }
    
    public InnocuousThread(Runnable task) {
        super(InnocuousThread.INNOCUOUSTHREADGROUP, task, newName());
        InnocuousThread.UNSAFE.putOrderedObject(this, InnocuousThread.INHERITEDACCESSCONTROLCONTEXT, InnocuousThread.ACC);
        this.eraseThreadLocals();
    }
    
    private InnocuousThread(ThreadGroup group, Runnable task, String name, ClassLoader classLoader) {
        super(group, task, name, 0L);
        InnocuousThread.UNSAFE.putOrderedObject(this, InnocuousThread.INHERITEDACCESSCONTROLCONTEXT, InnocuousThread.ACC);
        InnocuousThread.UNSAFE.putOrderedObject(this, InnocuousThread.CONTEXTCLASSLOADER, classLoader);
        this.eraseThreadLocals();
    }
    
    @Override
    public ClassLoader getContextClassLoader() {
        return ClassLoader.getSystemClassLoader();
    }
    
    @Override
    public void setUncaughtExceptionHandler(Thread.UncaughtExceptionHandler uncaughtExceptionHandler) {
    }
    
    @Override
    public void setContextClassLoader(ClassLoader classLoader) {
        throw new SecurityException("setContextClassLoader");
    }
    
    @Override
    public void run() {
        if (Thread.currentThread() == this && !this.hasRun) {
            this.hasRun = true;
            super.run();
        }
    }
    
    public void eraseThreadLocals() {
        InnocuousThread.UNSAFE.putObject(this, InnocuousThread.THREAD_LOCALS, null);
        InnocuousThread.UNSAFE.putObject(this, InnocuousThread.INHERITABLE_THREAD_LOCALS, null);
    }
    
    static {
        threadNumber = new AtomicInteger(1);
        try {
            ACC = new AccessControlContext(new ProtectionDomain[] { new ProtectionDomain(null, null) });
            UNSAFE = Unsafe.getUnsafe();
            Class<Object> clazz = Thread.class;
            Class<Object> clazz2 = ThreadGroup.class;
            THREAD_LOCALS = InnocuousThread.UNSAFE.objectFieldOffset(clazz.getDeclaredField("threadLocals"));
            INHERITABLE_THREAD_LOCALS = InnocuousThread.UNSAFE.objectFieldOffset(clazz.getDeclaredField("inheritableThreadLocals"));
            INHERITEDACCESSCONTROLCONTEXT = InnocuousThread.UNSAFE.objectFieldOffset(clazz.getDeclaredField("inheritedAccessControlContext"));
            CONTEXTCLASSLOADER = InnocuousThread.UNSAFE.objectFieldOffset(clazz.getDeclaredField("contextClassLoader"));
            long objectFieldOffset = InnocuousThread.UNSAFE.objectFieldOffset(clazz.getDeclaredField("group"));
            long objectFieldOffset2 = InnocuousThread.UNSAFE.objectFieldOffset(clazz2.getDeclaredField("parent"));
            final ThreadGroup threadGroup = (ThreadGroup)InnocuousThread.UNSAFE.getObject(Thread.currentThread(), objectFieldOffset);
            while (threadGroup != null) {
                ThreadGroup threadGroup2 = (ThreadGroup)InnocuousThread.UNSAFE.getObject(threadGroup, objectFieldOffset2);
                if (threadGroup2 == null) {
                    break;
                }
                threadGroup = threadGroup2;
            }
            INNOCUOUSTHREADGROUP = (ThreadGroup)AccessController.doPrivileged(new PrivilegedAction<ThreadGroup>() {
                @Override
                public ThreadGroup run() {
                    return new ThreadGroup(threadGroup, "InnocuousThreadGroup");
                }
            });
        }
        catch (Exception cause) {
            throw new Error(cause);
        }
    }
}
